from __future__ import annotations
from dataclasses import dataclass
import shapely


@dataclass(frozen=True)
class Region2D:
    """
    Simple data class to represent the bounding box for a region in 2D.

    :param x: the top left x-coordinate of the bounding box
    :param y: the top left y-coordinate of the bounding box
    :param width: the width of the bounding box
    :param height: the height of the bounding box
    :param z: the z-slice index of the bounding box
    :param t: the time point index of the bounding box
    """
    x: int = 0
    y: int = 0
    width: int = -1
    height: int = -1
    z: int = 0
    t: int = 0

    @property
    def geometry(self) -> shapely.Geometry:
        """
        A shapely geometry describing the x/y coordinates of the region.
        """
        return shapely.box(self.x, self.y, self.x+self.width, self.y+self.height)

    def scale_region(self, scale_factor: float) -> Region2D:
        """
        Scale the bounding box of this region on the x and y axis.

        :param scale_factor: the scale factor to apply to this region
        :returns: a Region2D scaled by the provided factor
        :raises ValueError: when scale_factor is 0
        """
        return self.downsample_region(1.0 / scale_factor)

    def downsample_region(self, downsample: float) -> Region2D:
        """
        Downsample the bounding box of this region on the x and y axis.

        This can be used to convert coordinates from (for example) the full image resolution 
        to a different pyramidal level.
        
        :param downsample: the downsample to apply to this region
        :returns: a Region2D downsampled by the provided factor
        :raises ValueError: when downsample is 0
        """
        if downsample == 1:
            return self

        if downsample == 0:
            raise ValueError('Downsample cannot be 0!')

        x = int(self.x / downsample)
        y = int(self.y / downsample)

        # Handle -1 for width & height, i.e. until the full image width
        if self.width == -1:
            x2 = x - 1
        else:
            x2 = int(round(self.x + self.width) / downsample)

        if self.height == -1:
            y2 = y - 1
        else:
            y2 = int(round(self.y + self.height) / downsample)

        return Region2D(x=x, y=y, width=x2 - x, height=y2 - y, z=self.z, t=self.t)



import numpy as np
import tempfile
import os
from enum import Enum
from py4j.java_gateway import JavaGateway, JavaObject
from urllib.parse import urlparse, unquote
from .image_server import ImageServer
from .metadata.image_metadata import ImageMetadata
from .metadata.image_shape import ImageShape
from .metadata.image_channel import ImageChannel
from .metadata.pixel_calibration import PixelCalibration, PixelLength
from .region_2d import Region2D
from .utils import base64_to_image, bytes_to_image
from ..qupath import qupath_gateway


class QuPathServer(ImageServer):
    """
    An ImageServer that communicates with a QuPath ImageServer through a Gateway.

    Closing this server won't close the underlying QuPath ImageServer.
    """

    def __init__(
        self,
        gateway: JavaGateway = None,
        qupath_server: JavaObject = None,
        pixel_access: str = 'base_64',
        **kwargs
    ):
        """
        :param gateway: the gateway between Python and QuPath. If not specified, the default gateway is used
        :param qupath_server: a Java object representing the QuPath ImageServer. If not specified, the currently
                              opened in QuPath ImageServer is used
        :param pixel_access: how to send pixel values from QuPath to Python. Can be 'base_64' to convert pixels to
                             base 64 encoded text, 'bytes' to to convert pixels to bytes, or 'temp_files' to use
                             temporary files
        :param resize_method: the resampling method to use when resizing the image for downsampling. Bicubic by default
        :raises ValueError: when pixel_access has an unexpected value
        """
        super().__init__(**kwargs)

        available_pixel_access = ['base_64', 'bytes', 'temp_files']
        if pixel_access not in available_pixel_access:
            raise ValueError(f'The provided pixel access ({pixel_access}) is not one of {str(available_pixel_access)}')

        self._gateway = qupath_gateway.get_default_gateway() if gateway is None else gateway
        self._qupath_server = qupath_gateway.get_current_image_data(gateway).getServer() if qupath_server is None else qupath_server
        self._pixel_access = pixel_access

    def close(self):
        pass

    def _build_metadata(self) -> ImageMetadata:
        qupath_metadata = self._qupath_server.getMetadata()

        return ImageMetadata(
            path=QuPathServer._find_qupath_server_path(self._qupath_server),
            name=qupath_metadata.getName(),
            shapes=tuple([
                ImageShape(
                    x=level.getWidth(),
                    y=level.getHeight(),
                    c=self._qupath_server.nChannels(),
                    z=self._qupath_server.nZSlices(),
                    t=self._qupath_server.nTimepoints()
                )
                for level in qupath_metadata.getLevels()
            ]),
            pixel_calibration=QuPathServer._find_qupath_server_pixel_calibration(self._qupath_server),
            is_rgb=self._qupath_server.isRGB(),
            dtype=np.dtype(self._qupath_server.getPixelType().toString().lower()),
            channels=tuple([ImageChannel(c.getName(), QuPathServer._unpack_color(c.getColor())) for c in qupath_metadata.getChannels()]),
            downsamples=tuple([d for d in self._qupath_server.getPreferredDownsamples()])
        )

    def _read_block(self, level: int, region: Region2D) -> np.ndarray:
        if level < 0:
            level = len(self.metadata.downsamples) + level
        downsample = self._qupath_server.getDownsampleForResolution(level)

        request = self._gateway.jvm.qupath.lib.regions.RegionRequest.createInstance(
            self._qupath_server.getPath(),
            downsample,
            int(round(region.x * downsample)),
            int(round(region.y * downsample)),
            int(round(region.width * downsample)),
            int(round(region.height * downsample)),
            region.z,
            region.t
        )

        if self._pixel_access == 'temp_files':
            temp_path = tempfile.mkstemp(prefix='qubalab-', suffix='.tif')[1]

            self._gateway.entry_point.writeImageRegion(self._qupath_server, request, temp_path)
            image = bytes_to_image(temp_path, self.metadata.is_rgb, ImageShape(region.width, region.height, c=self.metadata.n_channels))
            ## on Windows, this fails because the file handle is open elsewhere
            ## slightly bad manners to pollute tempfiles but should be insignificant
            if not os.name == "nt":
                os.remove(temp_path)
        else:
            format = 'png' if self.metadata.is_rgb else "imagej tiff"

            if self._pixel_access == 'bytes':
                image = bytes_to_image(
                    self._gateway.entry_point.getImageBytes(self._qupath_server, request, format),
                    self.metadata.is_rgb,
                    ImageShape(region.width, region.height, c=self.metadata.n_channels)
                )
            else:
                image = base64_to_image(
                    self._gateway.entry_point.getImageBase64(self._qupath_server, request, format),
                    self.metadata.is_rgb,
                    ImageShape(region.width, region.height, c=self.metadata.n_channels)
                )

        return image
    
    @staticmethod
    def _find_qupath_server_path(qupath_server: JavaObject) -> str:
        """
        Try to get the file path for a java object representing an ImageServer.
        This can be useful to get direct access to an image file, rather than via QuPath.
        """

        uris = tuple(str(u) for u in qupath_server.getURIs())
        if len(uris) == 1:
            parsed = urlparse(uris[0])
            if parsed.scheme == 'file':
                return unquote(parsed.path)
        return qupath_server.getPath()
    
    @staticmethod
    def _find_qupath_server_pixel_calibration(qupath_server: JavaObject) -> PixelCalibration:
        pixel_calibration = qupath_server.getPixelCalibration()

        if pixel_calibration.hasPixelSizeMicrons():
            return PixelCalibration(
                PixelLength.create_microns(pixel_calibration.getPixelWidthMicrons()),
                PixelLength.create_microns(pixel_calibration.getPixelHeightMicrons()),
                PixelLength.create_microns(pixel_calibration.getZSpacingMicrons()) if pixel_calibration.hasZSpacingMicrons() else PixelLength()
            )
        else:
            return PixelCalibration()
    
    @staticmethod
    def _unpack_color(rgb: int) -> tuple[float, float, float]:
        r = (rgb >> 16) & 255
        g = (rgb >> 8) & 255
        b = rgb & 255
        return r / 255.0, g / 255.0, b / 255.0



from matplotlib.axes import Axes
from typing import Union
import numpy as np
import shapely
from shapely.geometry import shape
from shapely import plotting
from geojson import GeoJSON
from ..objects.image_feature import ImageFeature


def plotImage(ax: Axes, image: np.ndarray, channel: int = 0, offset: tuple = None) -> Axes:
    """
    A utility function to display a NumPy array returned by an ImageServer.

    :param ax: the Axes to plot to
    :param image: a numpy array with dimensions (c, y, x). If it represents a RGB image, all channels will be displayed.
                  Otherwise, only the specified channel will be displayed
    :param channel: the channel to display if the image is not RGB
    :param offset: an offset [x_min, y_min, x_max, y_max] to apply to the image. Useful if the image should match some features.
                   None to not define any offset
    :return: the Axes where the image was plotted
    """
    
    # Move channel axis
    image = np.moveaxis(image, [0], [2])

    # Keep only specified channel if the image is not RGB
    if image.shape[2] != 3 or image.dtype != "uint8":
        image = image[..., channel]

    # If pixel values are float numbers, they can be greater than 0, while matplotlib only accepts
    # values between 0 and 1, so the array is normalized between 0 and 1
    if "float" in str(image.dtype):
        image /= image.max()

    if offset is not None and len(offset) == 4:
        # y_min and y_max are swapped because otherwise image data is flipped vertically
        ax.imshow(image, extent=(offset[0], offset[2], offset[3], offset[1]))
    else:
        ax.imshow(image)

    return ax


def plotImageFeatures(
    ax: Axes,
    image_features: list[ImageFeature],
    region: GeoJSON = None,
    default_color: tuple = (0.2, 0.8, 0.2),
    fill_opacity: float = 0.25
) -> Axes:
    """
    A utility function to display a image features. This will plot the object's geometries
    and nucleus geometries.
    
    :param ax: the Axes to plot to
    :param image_features: the image features to plot
    :param region: the region to plot. Features outside this region won't be plotted.
                   All features are plotted if this parameter is undefined. 
    :param default_color: the color to use for unclassified features
    :param fill_opacity: the opacity of the fill color. Set to 0 to not fill objects
    :return: the Axes where the image was plotted
    """
    envelope = None if region is None else shape(region).envelope

    for feature in image_features:
        if feature.color is not None:
            edge_color = tuple(c/255.0 for c in feature.color)
        elif feature.classification is not None:
            edge_color = tuple(c/255.0 for c in feature.classification.color)
        else:
            edge_color = default_color

        if fill_opacity > 0:
            face_color = edge_color + (fill_opacity,)
        else:
            face_color = None

        for roi in [getattr(feature, name, None) for name in ['geometry', 'nucleus_geometry']]:
            if roi is not None:
                geometry = shape(roi)
                if envelope is None or envelope.intersects(geometry.envelope):
                    _plot_geometry(ax, shape(roi), face_color, edge_color)

    ax.axis('equal')
    return ax


def _plot_geometry(ax: Axes, geometry: shapely.Geometry, face_color: tuple, edge_color: tuple):
    if (isinstance(geometry, Union[shapely.Point, shapely.MultiPoint])):
        plotting.plot_points(geometry, ax, color=edge_color)
    elif (isinstance(geometry, Union[shapely.LineString, shapely.LinearRing])):
        plotting.plot_line(geometry, ax, add_points=False, color=edge_color)
    elif (isinstance(geometry, Union[shapely.Polygon, shapely.MultiPolygon])):
        plotting.plot_polygon(geometry, ax, add_points=False, facecolor=face_color, edgecolor=edge_color, linewidth=2)
    elif (isinstance(geometry, shapely.MultiLineString)):
        for line in geometry.geoms:
            plotting.plot_line(line, ax, add_points=False, color=edge_color)
    elif (isinstance(geometry, shapely.GeometryCollection)):
        for g in geometry.geoms:
            _plot_geometry(ax, g, face_color, edge_color)



import numpy as np
import dask.array as da
from pathlib import Path
from dataclasses import astuple
import warnings
try:
    import openslide
except ImportError as e:
    warnings.warn(f'Unable to import OpenSlide, will try TiffSlide instead')
    import tiffslide as openslide
from .image_server import ImageServer
from .metadata.image_metadata import ImageMetadata
from .metadata.pixel_calibration import PixelCalibration, PixelLength
from .metadata.image_shape import ImageShape
from .region_2d import Region2D


class OpenSlideServer(ImageServer):
    """
    An image server that relies on OpenSlide (https://openslide.org/) to read RGB images.

    This server may only be able to detect the full resolution of a pyramidal image.

    OpenSlide provides some properties to define a rectangle bounding the non-empty region of the slide
    (see https://openslide.org/api/python/#standard-properties). If such properties are found, only this
    rectangle will be read (but note that this behaviour was not properly tested).
    """

    def __init__(self, path: str, strip_alpha=True, single_channel=False, limit_bounds=True, **kwargs):
        """
        :param path: the local path to the image to open
        :param strip_alpha: whether to strip the alpha channel from the image
        :param single_channel: whether to keep only the first channel of the image
        :param limit_bounds: whether to only consider a rectangle bounding the non-empty region of the slide,
                             if such rectangle is defined in the properties of the image
        :param resize_method: the resampling method to use when resizing the image for downsampling. Bicubic by default
        """
        super().__init__(**kwargs)
        self._reader = openslide.OpenSlide(path)
        self._path = path
        self._strip_alpha = strip_alpha
        self._single_channel = single_channel
        self._limit_bounds = limit_bounds

    def close(self):
        self._reader.close()

    def _build_metadata(self) -> ImageMetadata:
        n_channels = OpenSlideServer._get_n_channels(self._single_channel, self._strip_alpha)

        full_bounds = (0, 0) + self._reader.dimensions
        bounds = (self._reader.properties.get(openslide.PROPERTY_NAME_BOUNDS_X),
                  self._reader.properties.get(openslide.PROPERTY_NAME_BOUNDS_Y),
                  self._reader.properties.get(openslide.PROPERTY_NAME_BOUNDS_WIDTH),
                  self._reader.properties.get(openslide.PROPERTY_NAME_BOUNDS_HEIGHT))
        if self._limit_bounds and not any(v is None for v in bounds):
            self._bounds = bounds
            w = self._bounds[3]
            h = self._bounds[2]
            shapes = tuple(ImageShape(x=int(w / d), y=int(h / d), c=n_channels) for d in self._reader.level_downsamples)
        else:
            self._bounds = full_bounds
            shapes = tuple(ImageShape(x=d[0], y=d[1], c=n_channels) for d in self._reader.level_dimensions)

        return ImageMetadata(
            self._path,
            Path(self._path).name,
            shapes,
            OpenSlideServer._get_pixel_calibration(
                self._reader.properties.get(openslide.PROPERTY_NAME_MPP_X),
                self._reader.properties.get(openslide.PROPERTY_NAME_MPP_Y)
            ),
            True,
            np.uint8
        )

    def _read_block(self, level: int, region: Region2D) -> np.ndarray:
        x, y, width, height, z, t = astuple(region)
        assert z == 0, "OpenSlide can't read 3d images"
        assert t == 0, "OpenSlide can't read time-varying images"

        # x and y are provided in the coordinate space of the level, while openslide
        # expect coordinates in the level 0 reference frame
        x = int(x * self.metadata.downsamples[level])
        y = int(y * self.metadata.downsamples[level])

        # Map coordinates to bounding rectangle
        x += self._bounds[0]
        y += self._bounds[1]

        image = self._reader.read_region((x, y), level, (width, height))
        im = np.moveaxis(np.asarray(image), 2, 0)
        image.close()

        # Return image, stripping alpha/converting to single-channel if needed
        return im[:self.metadata.n_channels, :, :]

    @staticmethod
    def _get_n_channels(single_channel: bool, strip_alpha: bool):
        if single_channel:
            return 1
        elif strip_alpha:
            return 3
        else:
            return 4
        
    @staticmethod
    def _get_pixel_calibration(pixel_width: str, pixel_height: str):
        if pixel_width is not None and pixel_height is not None:
            return PixelCalibration(
                length_x=PixelLength.create_microns(float(pixel_width)),
                length_y=PixelLength.create_microns(float(pixel_height))
            )
        else:
            return PixelCalibration()



import geojson.geometry
import geojson
from typing import Union


class ImagePlane(dict):
    """
    An additional property of a geometry containing z and t indices.
    """

    def __init__(self, z: int = 0, t: int = 0):
        """
        :param z: the z-stack of the geometry
        :param t: the time point of the geometry
        """
        dict.__init__(self, z=z, t=t)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        raise AttributeError('ImagePlane is immutable')

    def __delattr__(self, name):
        raise AttributeError('ImagePlane is immutable')


def add_plane_to_geometry(
        geometry: Union[geojson.geometry.Geometry, geojson.Feature],
        z: int = None,
        t: int = None,
        preferred_geometry_key: str =None
) -> geojson.geometry.Geometry:
    """
    Create a GeoJSON Geometry object with an additional 'plane' property containing z and t indices.

    If indices are unspecified, they will be taken from the geometry's 'plane' property, if present,
    or else z or t attributes otherwise.
    If z and t values cannot be found in either of these locations, defaults (z=0 and t=0) will be
    used instead.

    The 'plane' property is immutable.

    :param geometry: a GeoJSON feature or geometry. It must not be a FeatureCollection
    :param z: the z-stack index this geometry should have
    :param t: the timepoint this geometry should have
    :param preferred_geometry_key: if the provided geometry is a feature, and this feature has a 'extra_geometries'
                                   property, then this parameter can be used as a key to retrieve the geometry to
                                   consider from the 'extra_geometries' dictionnary
    :return: a GeoJSON Geometry containing an additional 'plane' property containing z and t indices
    :raises ValueError: when the provided geometry is a FeatureCollection
    """
    if geometry is None:
        return None
    
    if geometry['type'] == 'Feature':
        feature = geometry
        if preferred_geometry_key is not None and 'extra_geometries' in feature.properties:
            geometry = feature.properties['extra_geometries'].get(preferred_geometry_key, feature.geometry)
        else:
            geometry = feature.geometry
    elif geometry['type'] == 'FeatureCollection':
        raise ValueError('Cannot convert FeatureCollection to single Geometry')

    if z is None or t is None:
        plane = getattr(geometry, 'plane', None)
        if z is None:
            z = getattr(plane, 'z', 0) if plane is not None else getattr(geometry, 'z', 0)
        if t is None:
            t = getattr(plane, 't', 0) if plane is not None else getattr(geometry, 't', 0)
    
    geometry = geojson.GeoJSON.to_instance(geometry, strict=False)
    geometry.plane = ImagePlane(z, t)
    return geometry



import numpy as np
import dask.array as da
import math
from pathlib import Path
from bioio import BioImage
from dataclasses import astuple
from .image_server import ImageServer
from .metadata.image_metadata import ImageMetadata
from .metadata.pixel_calibration import PixelCalibration, PixelLength
from .metadata.image_shape import ImageShape
from .region_2d import Region2D


class BioIOServer(ImageServer):
    """
    An ImageServer using BioIO (https://github.com/AllenCellModeling/BioIO).

    What this actually supports will depend upon how BioIO is installed.
    For example, it may provide Bio-Formats or CZI support... or it may not.

    Note that the BioIo library does not currently handle unit attachment, so the pixel unit
    given by this server will always be 'pixels'.
    
    Note that the BioIO library does not properly support pyramids, so you might only get the full
    resolution image when opening a pyramidal image.
    """

    def __init__(self, path: str, scene: int = 0, detect_resolutions=True, bioio_kwargs: dict[str, any] = {}, **kwargs):
        """
        :param path: the local path to the image to open
        :param scene: BioIO divides images into scene. This parameter specifies which scene to consider
        :param detect_resolutions: whether to look at all resolutions of the image (instead of just the full resolution)
        :param bioio_kwargs: any specific keyword arguments to pass down to the fsspec created filesystem handled by the BioIO reader
        :param resize_method: the resampling method to use when resizing the image for downsampling. Bicubic by default
        """
        super().__init__(**kwargs)
        self._path = path
        self._reader = BioImage(path, **bioio_kwargs)
        self._scene = scene
        self._detect_resolutions = detect_resolutions

    def level_to_dask(self, level: int = 0) -> da.Array:
        if level < 0 or level >= self.metadata.n_resolutions:
            raise ValueError("The provided level ({0}) is outside the valid range ([0, {1}])".format(level, self.metadata.n_resolutions - 1))

        axes = ("T" if self.metadata.n_timepoints > 1 else "") + \
            (("S" if "S" in self._reader.dims.order else "C") if self.metadata.n_channels > 1 else "") + \
            ("Z" if self.metadata.n_z_slices > 1 else "") + \
            "YX"

        self._reader.set_scene(self._reader.scenes[self._scene + level])
        return self._reader.get_image_dask_data(axes)

    def close(self):
        pass

    def _build_metadata(self) -> ImageMetadata:
        return ImageMetadata(
            self._path,
            Path(self._path).name,
            self._get_shapes(self._reader, self._scene) if self._detect_resolutions else (self._get_scene_shape(self._reader, self._reader.scenes[self._scene]),),
            self._get_pixel_calibration(self._reader, self._scene),
            self._is_rgb(self._reader, self._scene),
            np.dtype(self._reader.dtype)
        )

    def _read_block(self, level: int, region: Region2D) -> np.ndarray:
        x, y, width, height, z, t = astuple(region)
        
        self._reader.set_scene(self._reader.scenes[self._scene + level])
        axes = "TZ" + ("S" if "S" in self._reader.dims.order else "C") + "YX"

        return self._reader.get_image_dask_data(axes)[t, z, :, y:y + height, x:x + width].compute()

    @staticmethod
    def _get_shapes(reader: BioImage, scene: int) -> tuple[ImageShape, ...]:
        shapes = []
        for scene in reader.scenes[scene:]:
            shape = BioIOServer._get_scene_shape(reader, scene)

            if len(shapes) == 0 or BioIOServer._is_lower_resolution(shapes[-1], shape):
                shapes.append(shape)
            else:
                break
        return tuple(shapes)

    @staticmethod
    def _get_scene_shape(reader: BioImage, scene: int) -> ImageShape:
        reader.set_scene(scene)

        return ImageShape(
            x=reader.dims.X if 'X' in reader.dims.order else 1,
            y=reader.dims.Y if 'Y' in reader.dims.order else 1,
            z=reader.dims.Z if 'Z' in reader.dims.order else 1,
            c=reader.dims.S if 'S' in reader.dims.order else (reader.dims.C if 'C' in reader.dims.order else 1),
            t=reader.dims.T if 'T' in reader.dims.order else 1,
        )

    @staticmethod
    def _get_pixel_calibration(reader: BioImage, scene: int) -> PixelCalibration:
        reader.set_scene(scene)
        sizes = reader.physical_pixel_sizes

        if sizes.X or sizes.Y or sizes.Z:
            # The bioio library does not currently handle unit attachment, so the pixel unit is returned
            return PixelCalibration(
                PixelLength(sizes.X) if sizes.X is not None else PixelLength(),
                PixelLength(sizes.Y) if sizes.Y is not None else PixelLength(),
                PixelLength(sizes.Z) if sizes.Z is not None else PixelLength()
            )
        else:
            return PixelCalibration()
    
    @staticmethod
    def _is_rgb(reader: BioImage, scene: int) -> bool:
        reader.set_scene(scene)
        return ('S' in reader.dims.order and reader.dims.S in [3, 4]) or (reader.dtype == np.uint8 and reader.dims.C == 3)

    @staticmethod
    def _is_lower_resolution(base_shape: ImageShape, series_shape: ImageShape) -> bool:
        """
        Calculate if the series shape is a lower resolution than the base shape.

        This involves a bit of guesswork, but it's needed for so long as BioIO doesn't properly support pyramids.
        """
        if base_shape.z == series_shape.z and \
                base_shape.t == series_shape.t and \
                base_shape.c == series_shape.c:
            
            x_ratio = series_shape.x / base_shape.x
            y_ratio = series_shape.y / base_shape.y
            return x_ratio < 1.0 and math.isclose(x_ratio, y_ratio, rel_tol=0.01)
        else:
            return False




import numpy as np
import imageio
import base64
from .metadata.image_shape import ImageShape


def bytes_to_image(uri: {str, bytes}, is_rgb: bool, shape: ImageShape = None) -> np.ndarray:
    """
    Read the provided bytes (or URI pointing to a binary file) and convert them to a numpy array
    representing an image.

    :param uri: an URI pointing to a binary file, or an array of bytes. This must
                represent a 2-dimensionnal or 3-dimensionnal (with a channel axis) image
    :param is_rgb: whether the expected image should have the RGB format
    :param shape: the shape of the image to open. This will be used to reorder axes.
                  Can be None to not reorder axes
    :returns: a numpy array representing the input image. If shape is defined, this function
              will try to set its dimensions to (c, y, x), but this may not always be the case if two
              dimensions have the same number of elements
    """
    if is_rgb:
        image = imageio.v3.imread(uri)
    else:
        image = imageio.volread(uri, format="tiff")

    if shape is None:
        return image
    else:
        return _reorder_axes(image, shape)


def base64_to_image(data: str, is_rgb: bool, shape: ImageShape = None) -> np.ndarray:
    """
    Read the provided string with the Base64 format and convert it to a numpy array
    representing an image.

    :param uri: a text with the Base64 format. This must represent a 2-dimensionnal
                or 3-dimensionnal (with a channel axis) image
    :param is_rgb: whether the expected image should have the RGB format
    :param shape: the shape of the image to open. This will be used to reorder axes.
                  Can be None to not reorder axes
    :returns: a numpy array representing the input image. If shape is defined, this function
              will try to set its dimensions to (c, y, x), but this may not always be the case if two
              dimensions have the same number of elements
    """
    return bytes_to_image(base64.b64decode(data), is_rgb, shape)


def _reorder_axes(image: np.ndarray, shape: ImageShape):
    """
    Reorder the axes of the provided image to (c, y, x).

    :param image: the image to reorder
    :param shape: the desired shape of the image
    :param shape: the shape of the image to open. This will be used to reorder axes
    :returns: a numpy array representing the input image. This function will try to set
              its dimensions to (c, y, x), but this may not be the case if two
              dimensions have the same number of elements
    """
    x_axis = -1
    y_axis = -1
    c_axis = -1
    for axis, number_of_elements_on_axis in enumerate(image.shape):
        if c_axis == -1 and number_of_elements_on_axis == shape.c:
            c_axis = axis
        elif y_axis == -1 and number_of_elements_on_axis == shape.y:
            y_axis = axis
        elif x_axis == -1 and number_of_elements_on_axis == shape.x:
            x_axis = axis

    if x_axis != -1 and y_axis != -1 and c_axis != -1:
        return np.moveaxis(image, [c_axis, y_axis, x_axis], [0, 1, 2]).copy()
    else:
        return image



import logging
import math
import numpy as np
from .image_channel import ImageChannel
from .image_shape import ImageShape
from .pixel_calibration import PixelCalibration


class ImageMetadata:
    """
    Simple class to store core metadata for a pyramidal image.
    """

    def __init__(
        self,
        path: str,
        name: str,
        shapes: tuple[ImageShape, ...],
        pixel_calibration: PixelCalibration,
        is_rgb: bool,
        dtype: np.dtype,
        channels: tuple[ImageChannel, ...] = None,
        downsamples = None
    ):
        """
        :param path: the local path to the image
        :param name: the image name
        :param shapes: the image shape, for each resolution of the image
        :param pixel_calibration: the pixel calibration information of the image
        :param is_rgb: whether pixels of the image are stored with the RGB format
        :param dtype: the type of the pixel values
        :param _channels: the channels of the image (optional)
        :param _downsamples: the downsamples of the image (optional)
        """
        self.path = path
        self.name = name
        self.shapes = shapes
        self.pixel_calibration = pixel_calibration
        self.is_rgb = is_rgb
        self.dtype = dtype
        self._channels = channels
        self._downsamples = downsamples

    _DEFAULT_CHANNEL_SINGLE = (
        ImageChannel(name='Single channel', color=(1, 1, 1)),
    )
    _DEFAULT_CHANNEL_RGB = (
        ImageChannel(name='Red', color=(1, 0, 0)),
        ImageChannel(name='Green', color=(0, 1, 0)),
        ImageChannel(name='Green', color=(0, 0, 1)),
    )
    _DEFAULT_CHANNEL_TWO = (
        ImageChannel(name='Channel 1', color=(1, 0, 1)),
        ImageChannel(name='Channel 2', color=(0, 1, 0))
    )
    _DEFAULT_CHANNEL_COLORS = (
        (0, 1, 1),
        (1, 1, 0),
        (1, 0, 1),
        (1, 0, 0),
        (0, 1, 0),
        (0, 0, 1)
    )

    @property
    def shape(self) -> ImageShape:
        """
        The dimensions of the full-resolution image.
        """
        return self.shapes[0]
    
    @property
    def width(self) -> int:
        """
        The width of the full-resolution image.
        """
        return self.shape.x

    @property
    def height(self) -> int:
        """
        The height of the full-resolution image.
        """
        return self.shape.y

    @property
    def n_channels(self) -> int:
        """
        The number of channels of the image.
        """
        return self.shape.c

    @property
    def n_timepoints(self) -> int:
        """
        The number of time points of the image.
        """
        return self.shape.t

    @property
    def n_z_slices(self) -> int:
        """
        The number of z-slices of the image.
        """
        return self.shape.z

    @property
    def n_resolutions(self) -> int:
        """
        The number of resolutions of the image.
        """
        return len(self.shapes)

    @property
    def downsamples(self) -> tuple[float, ...]:
        """
        The downsamples of the image.
        """
        if self._downsamples is None:
            self._downsamples = tuple(self._estimate_downsample(self.shape, s) for s in self.shapes)
        return self._downsamples

    @property
    def channels(self) -> tuple[ImageChannel, ...]:
        """
        The channels of the image.
        """
        if self._channels is None:
            if self.is_rgb:
                self._channels = self._DEFAULT_CHANNEL_RGB
            else:
                if self.n_channels == 1:
                    self._channels = self._DEFAULT_CHANNEL_SINGLE
                elif self.n_channels == 2:
                    self._channels = self._DEFAULT_CHANNEL_TWO
                else:
                    self._channels = [
                        ImageChannel(
                            f'Channel {ii + 1}',
                            self._DEFAULT_CHANNEL_COLORS[ii % len(self._DEFAULT_CHANNEL_COLORS)]
                        ) for ii in range(self.n_channels)
                    ]
        return self._channels
    
    def __eq__(self, other):
        if isinstance(other, ImageMetadata):
            return self.path == other.path and \
                self.path == other.path and self.name == other.name and self.shapes == other.shapes \
                and self.pixel_calibration == other.pixel_calibration and self.is_rgb == other.is_rgb \
                and self.dtype == other.dtype and self.channels == other.channels and self.downsamples == other.downsamples
        else:
            return False
    
    def _estimate_downsample(self, higher_resolution_shape: ImageShape, lower_resolution_shape: ImageShape) -> float:
        """
        Estimate the downsample factor between a higher resolution and a lower resolution ImageShape.
        
        This is used to prefer values like 4 rather than 4.000345, which arise due to resolutions having to have
        integer pixel dimensions.
        The downsample is computed between the width and the heights of the shapes.
        :param higher_resolution_shape: a higher resolution shape
        :param lower_resolution_shape: a lower resolution shape
        :return: the downsample between the higher and the lower resolution shape
        """

        dx = higher_resolution_shape.x / lower_resolution_shape.x
        dy = higher_resolution_shape.y / lower_resolution_shape.y
        downsample = (dx + dy) / 2.0
        downsample_round = round(downsample)

        if (
            self._possible_downsample(higher_resolution_shape.x, lower_resolution_shape.x, downsample_round) and
            self._possible_downsample(higher_resolution_shape.y, lower_resolution_shape.y, downsample_round)
        ):
            if downsample != downsample_round:
                logging.debug(f'Returning rounded downsample value {downsample_round} instead of {downsample}')
            return downsample_round
        else:
            return downsample

    def _possible_downsample(self, higher_resolution_value: int, lower_resolution_value: int, downsample: float) -> bool:
        """
        Determine if an image dimension is what you'd expect after downsampling and applying floor or ceil.

        :param higher_resolution_value: a higher resolution value to downsample
        :param lower_resolution_value: a lower resolution value to compare to the downsampled higher resolution value
        :param downsample: the downsample to apply to the higher resolution value
        :return: whether the downsampled higher resolution value corresponds to the lower resolution value
        """
        return (math.floor(higher_resolution_value / downsample) == lower_resolution_value or
                math.ceil(higher_resolution_value / downsample) == lower_resolution_value)



from .image_server import ImageServer
from .metadata.image_metadata import ImageMetadata


class WrappedImageServer(ImageServer):
    """
    Abstract class for an ImageServer that wraps another ImageServer,
    e.g. to transform the image in some way.
    
    Closing this server will close the wrapped server.
    The metadata of this server is equivalent to the metadata of the wrapped server.
    """

    def __init__(self, base_server: ImageServer, **kwargs):
        """
        :param base_server: the server to wrap
        :param resize_method: the resampling method to use when resizing the image for downsampling. Bicubic by default
        """
        super().__init__(**kwargs)
        self._base_server = base_server

    @property
    def base_server(self) -> ImageServer:
        """
        Get the wrapped server.
        """
        return self._base_server
    
    def close(self):
        self._base_server.close()

    def _build_metadata(self) -> ImageMetadata:
        return self._base_server.metadata



import numpy as np
from py4j.java_gateway import JavaGateway, GatewayParameters, JavaObject
from py4j.protocol import Py4JNetworkError
from enum import Enum
from typing import Union
import geojson
from ..images import utils
from ..objects.image_feature import ImageFeature
from ..objects.object_type import ObjectType
from ..objects.geojson import geojson_features_from_string


_default_gateway = None


def create_gateway(auto_convert=True, auth_token=None, port=25333, set_as_default=True, **kwargs) -> JavaGateway:
    """
    Create a new JavaGateway to communicate with QuPath.

    This requires launching QuPath and activating Py4J from there first.

    :param auto_convert: if True, the gateway will try to automatically convert Python objects like sequences and maps to Java Objects
    :param auth_token: the authentication token to use to connect. Can be None if no authentication is used
    :param port: the port to use to connect
    :param set_as_default: whether to set the created gateway as the default gateway
    :param kwargs: additional arguments to give to the JavaGateway constructor
    :returns: the created gateway
    :raises RuntimeError: if the connection to QuPath couldn't be established
    """
    if auth_token is None:
        gateway_parameters = GatewayParameters(port=port)
    else:
        gateway_parameters = GatewayParameters(auth_token=auth_token, port=port)
    gateway = JavaGateway(
        auto_convert=auto_convert,
        gateway_parameters=gateway_parameters,
        **kwargs
    )

    try:
        # This will fail if QuPath is not running with a Py4J gateway
        qupath = gateway.entry_point.getQuPath()
        assert qupath is not None
    except (Py4JNetworkError, AssertionError) as err:
        raise RuntimeError('Could not connect to QuPath - is it running, and you have opened a Py4J gateway?') from err
    
    if set_as_default:
        set_default_gateway(gateway)
    return gateway


def set_default_gateway(gateway: JavaGateway):
    """
    Set the default JavaGateway to use if one is not otherwise specified.

    This default gateway will be used if no gateway is provided in some functions of this file.

    :param gateway: the gateway that should be the default gateway
    """
    global _default_gateway
    _default_gateway = gateway


def get_default_gateway() -> JavaGateway:
    """
    Return the default gateway. It will be created if it doesn't already exist.

    :returns: the default gateway
    """
    global _default_gateway
    if _default_gateway is None:
        _default_gateway = create_gateway()
    return _default_gateway


def get_current_image_data(gateway: JavaGateway = None) -> JavaObject:
    """
    Get the current ImageData opened in QuPath through the provided gateway.

    :param gateway: the gateway to use. Can be None to use the default gateway
    :returns: the current ImageData opened in QuPath
    """
    gateway = get_default_gateway() if gateway is None else gateway
    return gateway.entry_point.getQuPath().getImageData()


def get_project(gateway: JavaGateway = None) -> JavaObject:
    """
    Return the currently opened QuPath project.

    :param gateway: the gateway to use. Can be None to use the default gateway
    :return: a Java Object representing the currently opened QuPath project
    """
    gateway = get_default_gateway() if gateway is None else gateway
    return gateway.entry_point.getProject()


def create_snapshot(gateway: JavaGateway = None, snapshot_type: str = 'qupath') -> np.ndarray:
    """
    Create and return a snapshot of QuPath. 

    :param gateway: the gateway to use. Can be None to use the default gateway
    :param snapshot_type: what to include in the snapshot. 'qupath' for the entire qupath window or
                          'viewer' for only the viewer
    :returns: a numpy array with dimensions (y, x, c) representing an RGB image
    :raises ValueError: if the snapshot type was not recognized
    """
    gateway = get_default_gateway() if gateway is None else gateway
    qp = gateway.entry_point

    if snapshot_type == 'qupath':
        image = qp.snapshotBase64(qp.getQuPath())
    elif snapshot_type == 'viewer':
        image = qp.snapshotBase64(qp.getCurrentViewer())
    else:
        raise ValueError(f'Unknown snapshot_type {snapshot_type}')
        
    return utils.base64_to_image(image, True)


def get_objects(
    image_data: JavaObject = None,
    gateway: JavaGateway = None,
    object_type: ObjectType = None,
    converter: str = None
) -> Union[list[geojson.Feature], list[ImageFeature], list[JavaObject]]:
    """
    Get the objects (e.g. detections, annotations) of the current or specified image in QuPath.

    :param image_data: the image_data to retrieve objects from. Can be None to use the current ImageData opened in QuPath
    :param gateway: the gateway to use. Can be None to use the default gateway
    :param object_type: the type of object to get. Can be None to get all objects except the root
    :param converter: can be 'geojson' to get an extended GeoJSON feature represented as a QuBaLab 'ImageObject',
                      or 'simple_feature' to get a regular GeoJSON 'Feature'. By default, the Py4J representation of
                      the QuPath Java object is returned.
    :return: the list of objects of the specified image. The type of the objects depends on the converted parameter
    """
    gateway = get_default_gateway() if gateway is None else gateway
    image_data = get_current_image_data(gateway) if image_data is None else image_data
    hierarchy = image_data.getHierarchy()

    match object_type:
        case None:
            path_objects = hierarchy.getAllObjects(False)
        case ObjectType.ANNOTATION:
            path_objects = hierarchy.getAnnotationObjects()
        case ObjectType.DETECTION:
            path_objects = hierarchy.getDetectionObjects()
        case ObjectType.TILE:
            path_objects = hierarchy.getTileObjects()
        case ObjectType.CELL:
            path_objects = hierarchy.getCellObjects()
        case ObjectType.TMA_CORE:
            tma_grid = hierarchy.getTMAGrid()
            path_objects = [] if tma_grid is None else tma_grid.getTMACoreList()

    if converter == 'geojson' or converter == 'simple_feature':
        # Use toFeatureCollections for performance and to avoid string length troubles
        features = []
        for feature_collection in gateway.entry_point.toFeatureCollections(path_objects, 1000):
            features.extend(geojson_features_from_string(feature_collection, parse_constant=None))

        if converter == 'simple_feature':
            return features
        else:
            return [ImageFeature.create_from_feature(f) for f in features]
    else:
        return path_objects


def count_objects(image_data: JavaObject = None, gateway: JavaGateway = None, object_type: ObjectType = None,) -> int:
    """
    Get a count of all objects in the current or specified image.

    Since requesting the objects can be slow, this can be used to check if a reasonable
    number of objects is available.

    :param image_data: the image_data to retrieve objects from. Can be None to use the current ImageData opened in QuPath
    :param gateway: the gateway to use. Can be None to use the default gateway
    :param object_type: the type of object to get. Can be None to get all objects except the root
    :return: the number of such objects in the current or specified image
    """
    return len(get_objects(image_data, gateway, object_type))


def add_objects(features: Union[list[geojson.Feature], geojson.Feature], image_data: JavaObject = None, gateway: JavaGateway = None):
    """
    Add the provided features to the current or provided ImageData.

    :param features: the features to add. Can be a list or a single Feature
    :param image_data: the image_data to add features to. Can be None to use the current ImageData opened in QuPath
    :param gateway: the gateway to use. Can be None to use the default gateway
    """
    gateway = get_default_gateway() if gateway is None else gateway
    image_data = get_current_image_data(gateway) if image_data is None else image_data

    if isinstance(features, geojson.Feature):
        features = list(features)
    
    features_json = geojson.dumps(features, allow_nan=True)
    image_data.getHierarchy().addObjects(gateway.entry_point.toPathObjects(features_json))


def delete_objects(image_data: JavaObject = None, gateway: JavaGateway = None, object_type: ObjectType = None):
    """
    Delete all specified objects (e.g. annotations, detections) from the current or provided ImageData.

    :param image_data: the image_data to remove the objects from. Can be None to use the current ImageData opened in QuPath
    :param gateway: the gateway to use. Can be None to use the default gateway
    :param object_type: the type of object to remove. Can be None to remove all objects
    """
    gateway = get_default_gateway() if gateway is None else gateway
    image_data = get_current_image_data(gateway) if image_data is None else image_data

    if image_data is not None:
        image_data.getHierarchy().removeObjects(get_objects(image_data, gateway, object_type), True)
        image_data.getHierarchy().getSelectionModel().clearSelection()


def refresh_qupath(gateway: JavaGateway = None):
    """
    Update the current QuPath interface.
    
    This is sometimes needed to update the QuPath window when some changes are
    made to hierarchy.

    :param gateway: the gateway to use. Can be None to use the default gateway
    """
    gateway = get_default_gateway() if gateway is None else gateway
    gateway.entry_point.fireHierarchyUpdate()



from __future__ import annotations
from dataclasses import dataclass


@dataclass
class ImageShape:
    """
    Simple data class to store an image shape.

    :param x: the image width
    :param y: the image height
    :param t: the number of time points
    :param c: the number of channels
    :param z: the number of z-stacks
    """
    x: int
    y: int
    t: int = 1
    c: int = 1
    z: int = 1

    def from_tczyx(*args) -> ImageShape:
        """
        Create an ImageShape from a list of arguments.
        
        :param args: image width, image height, number of time points, number of channels, number of z-stacks
                     (in that order)
        :returns: an ImageShape corresponding to the arguments
        :raises IndexeError: when there are less than five arguments
        """
        return ImageShape(t=args[0], c=args[1], z=args[2], y=args[3], x=args[4])

    def as_tuple(self, dims: str = 'tczyx') -> tuple:
        """
        Return a tuple describing this ImageShape.

        :param dims: the format the resulting tuple should have. Each character must be one of 'tczyx'
        :returns: a tuple describing this ImageShape with the specified format
        :raises AttributeError: when a character of dims is not in 'tczyx'
        """
        return tuple(self.__getattribute__(d) for d in dims)



from __future__ import annotations
import geojson
import uuid
import math
import numpy as np
import geojson.geometry
from typing import Union, Any
import rasterio
import rasterio.features
import shapely
from .object_type import ObjectType
from .classification import Classification
from .geometry import add_plane_to_geometry


_NUCLEUS_GEOMETRY_KEY = 'nucleus'


class ImageFeature(geojson.Feature):
    """
    GeoJSON Feature with additional properties for image objects.

    The added properties are:

    - A classification (defined by a name and a color).

    - A name.

    - A list of measurements.

    - A type of QuPath object (e.g. detection, annotation).

    - A color.

    - Additional geometries.
    
    - And any other property.
    """

    def __init__(
        self,
        geometry: geojson.geometry.Geometry,
        classification: Union[Classification, dict] = None,
        name: str = None,
        measurements: dict[str, float] = None,
        object_type: ObjectType = ObjectType.ANNOTATION,
        color: tuple[int, int, int] = None,
        extra_geometries: dict[str, geojson.geometry.Geometry] = None,
        id: Union[str, int, uuid.UUID] = None,
        extra_properties: dict[str, Any] = None
    ):
        """
        Except from the geometry and id parameters, all parameters of this
        constructor will be added to the list of properties of this feature
        (if provided).
        
        :param geometry: the geometry of the feature
        :param classification: the classification of this feature, or a dictionnary with the
                               'name' and 'color' properties defining respectively a string
                               and a 3-long int tuple with values between 0 and 255
        :param name: the name of this feature
        :param measurements: a dictionnary containing measurements. Measurements
                             with NaN values will not be added
        :param object_type: the type of QuPath object this feature represents
        :param color: the color of this feature
        :param extra_geometries: a dictionnary containing additional geometries
                                 that represent this feature
        :param id: the ID of the feature. If not provided, an UUID will be generated
        :param extra_properties: a dictionnary of additional properties to add
        """
        props = {}
        if classification is not None:
            if isinstance(classification, Classification):
                props['classification'] = {
                    "name": classification.name,
                    "color": classification.color
                }
            else:
                props['classification'] = {
                    "name": classification.get('name'),
                    "color": classification.get('color')
                }
        if name is not None:
            props['name'] = name
        if measurements is not None:
            props['measurements'] = ImageFeature._remove_NaN_values_from_measurements(measurements)
        if object_type is not None:
            props['object_type'] = object_type.name
        if color is not None:
            props['color'] = color
        if extra_geometries is not None:
            props['extra_geometries'] = {k: add_plane_to_geometry(v) for k, v in extra_geometries.items()}
        if extra_properties is not None:
            props.update(extra_properties)
        
        super().__init__(
            geometry=add_plane_to_geometry(geometry),
            properties=props,
            id=ImageFeature._to_id_string(id)
        )
        self['type'] = 'Feature'
        
    
    @classmethod
    def create_from_feature(cls, feature: geojson.Feature) -> ImageFeature:
        """
        Create an ImageFeature from a GeoJSON feature.

        The ImageFeature properties will be searched in the provided
        feature and in the properties of the provided feature.
        
        :param feature: the feature to convert to an ImageFeature
        :return: an ImageFeature corresponding to the provided feature
        """
        geometry = cls._find_property(feature, 'geometry')
        plane = cls._find_property(feature, 'plane')
        if plane is not None:
            geometry = add_plane_to_geometry(geometry, z=getattr(plane, 'z', None), t=getattr(plane, 't', None))

        object_type_property = cls._find_property(feature, 'object_type')
        if object_type_property is None:
            object_type_property = cls._find_property(feature, 'objectType')
        object_type = next((o for o in ObjectType if o.name.lower() == str(object_type_property).lower()), None)

        args = dict(
            geometry=geometry,
            id=cls._find_property(feature, 'id'),
            classification=cls._find_property(feature, 'classification'),
            name=cls._find_property(feature, 'name'),
            color=cls._find_property(feature, 'color'),
            measurements=cls._find_property(feature, 'measurements'),
            object_type=object_type,
        )

        nucleus_geometry = cls._find_property(feature, 'nucleusGeometry')
        if nucleus_geometry is not None:
            if plane is not None:
                nucleus_geometry = add_plane_to_geometry(nucleus_geometry, z=getattr(plane, 'z', None), t=getattr(plane, 't', None))
            args['extra_geometries'] = dict(nucleus=nucleus_geometry)

        args['extra_properties'] = {k: v for k, v in feature['properties'].items() if k not in args and v is not None}
        return cls(**args)
    
    @classmethod
    def create_from_label_image(
        cls,
        input_image: np.ndarray,
        object_type: ObjectType = ObjectType.ANNOTATION,
        connectivity: int = 4,
        scale: float = 1.0,
        include_labels = False,
        classification_names: Union[str, dict[int, str]] = None
    ) -> list[ImageFeature]:
        """
        Create a list of ImageFeatures from a binary or labeled image.

        The created geometries will be polygons, even when representing points or line.

        :param input_image: a 2-dimensionnal binary (with a boolean type) or labeled
                            (with a uint8 type) image containing the features to create.
                            If a binary image is given, all True pixel values will be
                            considered as potential features. If a labeled image is given,
                            all pixel values greater than 0 will be considered as potential features
        :param object_type: the type of object to create
        :param connectivity: the pixel connectivity for grouping pixels into features (4 or 8)
        :param scale: a scale value to apply to the shapes
        :param include_labels: whether to include a 'Label' measurement in the created features
        :param classification_names: if str, the name of the classification to apply to all features.
                                     if dict, a dictionnary mapping a label to a classification name
        :return: a list of image features representing polygons present in the input image
        """
        features = []

        if input_image.dtype == bool:
            mask = input_image
            input_image = input_image.astype(np.uint8)
        else:
            mask = input_image > 0

        transform = rasterio.transform.Affine.scale(scale)

        existing_features = {}
        for geometry, label in rasterio.features.shapes(input_image, mask=mask, connectivity=connectivity, transform=transform):
            if label in existing_features:
                existing_features[label]['geometry'] = shapely.geometry.shape(geometry).union(
                    shapely.geometry.shape(existing_features[label]['geometry'])
                )
            else:
                if isinstance(classification_names, str):
                    classification_name = classification_names
                elif isinstance(classification_names, dict) and int(label) in classification_names:
                    classification_name = classification_names[int(label)]
                else:
                    classification_name = None

                feature = cls(
                    geometry=geometry,
                    classification=Classification.get_cached_classification(classification_name),
                    measurements={'Label': float(label)} if include_labels else None,
                    object_type=object_type
                )

                existing_features[label] = feature
                features.append(feature)

        # Ensure we have GeoJSON-compatible geometries
        for feature in features:
            feature['geometry'] = geojson.mapping.to_mapping(feature['geometry'])

        return features

    @property
    def classification(self) -> Classification:
        """
        The classification of this feature (or None if not defined).
        """
        if "classification" in self.properties:
            return Classification(self.properties['classification'].get('name'), self.properties['classification'].get('color'))
        else:
            return None

    @property
    def name(self) -> str:
        """
        The name of this feature (or None if not defined).
        """
        return self.properties.get('name')

    @property
    def measurements(self) -> dict[str, float]:
        """
        The measurements of this feature.
        """
        measurements = self.properties.get('measurements')
        if measurements is None:
            measurements = {}
            self.properties['measurements'] = measurements
        return measurements

    @property
    def object_type(self) -> ObjectType:
        """
        The QuPath object type (e.g. detection, annotation) this feature represents
        or None if the object type doesn't exist or is not recognised.
        """
        return next((o for o in ObjectType if o.name.lower() == str(self.properties['object_type']).lower()), None)

    @property
    def is_detection(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is a detection, cell, or tile.
        """
        return self.object_type in [ObjectType.DETECTION, ObjectType.CELL, ObjectType.TILE]

    @property
    def is_cell(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is a cell.
        """
        return self.object_type == ObjectType.CELL

    @property
    def is_tile(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is a tile.
        """
        return self.object_type == ObjectType.TILE

    @property
    def is_annotation(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is an annotation.
        """
        return self.object_type == ObjectType.ANNOTATION

    @property
    def color(self) -> tuple[int, int, int]:
        """
        The color of this feature (or None if not defined).
        """
        return self.properties.get('color')

    @property
    def nucleus_geometry(self) -> geojson.geometry.Geometry:
        """
        The nucleus geometry of this feature (or None if not defined).
        It can be defined when passed as an extra_geometry with the 'nucleus'
        key when creating an ImageFeature, by defining the 'nucleus_geometry'
        property of an ImageFeature, or when passed as a 'nucleusGeometry'
        property when creating an ImageFeature from a GeoJSON feature.
        """
        extra = self.properties.get('extra_geometries')
        if extra is not None:
            return extra.get(_NUCLEUS_GEOMETRY_KEY)
        return None

    def __setattr__(self, name, value):
        if name == 'classification':
            if isinstance(value, Classification):
                self.properties['classification'] = {
                    "name": value.name,
                    "color": value.color
                }
            else:
                self.properties['classification'] = value
        elif name == 'name':
            self.properties['name'] = value
        elif name == 'measurements':
            self.properties['measurements'] = ImageFeature._remove_NaN_values_from_measurements(value)
        elif name == 'object_type':
            if isinstance(value, str):
                self.properties['object_type'] = value
            elif isinstance(value, ObjectType):
                self.properties['object_type'] = value.name
        elif name == 'color':
            if len(value) != 3:
                raise ValueError('Color must be a tuple of length 3')
            rgb = tuple(ImageFeature._validate_rgb_value(v) for v in value)
            self.properties['color'] = rgb
        elif name == 'nucleus_geometry':
            if 'extra_geometries' not in self.properties:
                self.properties['extra_geometries'] = {}
            self.properties['extra_geometries'][_NUCLEUS_GEOMETRY_KEY] = add_plane_to_geometry(value)
        else:
            super().__setattr__(name, value)
    
    @staticmethod
    def _remove_NaN_values_from_measurements(measurements: dict[str, float]) -> dict[str, float]:
        return {
            k: float(v) for k, v in measurements.items()
            if isinstance(k, str) and isinstance(v, (int, float)) and not math.isnan(v)
        }
    
    @staticmethod
    def _to_id_string(object_id: Union[int, str, uuid.UUID]) -> str:
        if object_id is None:
            return str(uuid.uuid4())
        elif isinstance(object_id, str) or isinstance(object_id, int):
            return object_id
        else:
            return str(object_id)
    
    @staticmethod
    def _find_property(feature: geojson.Feature, property_name: str):
        if property_name in feature:
            return feature[property_name]
        elif 'properties' in feature and property_name in feature['properties']:
            return feature['properties'][property_name]
        else:
            return None

    @staticmethod
    def _validate_rgb_value(value: Union[int, float]) -> int:
        if isinstance(value, float):
            value = int(math.round(value * 255))
        if isinstance(value, int):
            if value >= 0 and value <= 255:
                return value
        raise ValueError('Color value must be an int between 0 and 255, or a float between 0 and 1')



from typing import Union
from PIL import Image, ImageDraw
import shapely


def draw_geometry(image_size: tuple[int, int], drawing_context: ImageDraw.Draw, geometry: shapely.Geometry, value: int):
    """
    Draw the provided Shapely geometry with the provided drawing context using the provided value.

    :param image_size: the size of the image to draw on
    :param drawing_context: the drawing context to use
    :param geometry: the geometry to draw
    :param value: the value to use when drawing
    """
    if (isinstance(geometry, shapely.Point)):
        _draw_point(drawing_context, geometry, value)
    elif (isinstance(geometry, Union[shapely.LineString, shapely.LinearRing])):
        _draw_line(drawing_context, geometry, value)
    elif (isinstance(geometry, shapely.Polygon)):
        _draw_polygon(image_size, drawing_context, geometry, value)
    elif (isinstance(geometry, shapely.MultiPoint)):
        for point in geometry.geoms:
            _draw_point(drawing_context, point, value)
    elif (isinstance(geometry, shapely.MultiLineString)):
        for line in geometry.geoms:
            _draw_line(drawing_context, line, value)
    elif (isinstance(geometry, shapely.MultiPolygon)):
        for polygon in geometry.geoms:
            _draw_polygon(image_size, drawing_context, polygon, value)
    elif (isinstance(geometry, shapely.GeometryCollection)):
        for g in geometry.geoms:
            draw_geometry(image_size, drawing_context, g, value)

def _draw_point(drawing_context: ImageDraw, point: shapely.Point, value: int):
    drawing_context.point([point.x, point.y], value)

def _draw_line(drawing_context: ImageDraw, line: Union[shapely.LineString, shapely.LinearRing], value: int):
    drawing_context.line(line.coords, value)

def _draw_polygon(image_size: tuple[int, int], drawing_context: ImageDraw, polygon: shapely.Polygon, value: int):
    # If we have holes, we risk overpainting existing pixels with the background color
    # For that reason we need to create a separate binary image and then copy the values
    if len(polygon.interiors) > 0:
        bitmap = Image.new('1', image_size)
        bitmap_draw = ImageDraw.Draw(bitmap)

        bitmap_draw.polygon(polygon.exterior.coords, 1)
        for interior in polygon.interiors:
            bitmap_draw.polygon(interior.coords, 0)

        drawing_context.bitmap((0, 0), bitmap, value)
    else:
        drawing_context.polygon(polygon.exterior.coords, value)



from __future__ import annotations
import random


class Classification(object):
    """
    Simple class to store the name and color of a classification.
    """
    _cached_classifications = {}

    def __init__(self, name: str, color: tuple[int, int, int] = None):
        """
        :param name: the name of the classification
        :param color: the RGB color (each component between 0 and 255) of the classification. Can be None to use a random color
        """
        self._name = name
        self._color = [random.randint(0, 255) for _ in range(3)] if color is None else color

    @property
    def name(self) -> str:
        """
        The name of the classification.
        """
        return self._name

    @property
    def color(self) -> tuple[int, int, int]:
        """
        The color of the classification.
        """
        return self._color
    
    @staticmethod
    def get_cached_classification(name: str, color: tuple[int, int, int] = None) -> Classification:
        """
        Return a classification by looking at an internal cache.
        
        If no classification with the provided name is present in the cache, a
        new classification is created and the cache is updated.

        This is useful if you want to avoid creating multiple classifications with the
        same name and use only one instead.

        :param name: the name of the classification (can be None)
        :param color: the RGB color (each component between 0 and 255) of the classification.
                      Can be None to use a random color. This is only used if the cache doesn't
                      already contain a classification with the provided name
        :return: a classification with the provided name, but not always with the provided color
                 if a classification with the same name already existed in the cache. If the provided
                 name is None, None is also returned
        """
        if name is None:
            return None
        
        classification = Classification._cached_classifications.get(name)
        if classification is None:
            classification = Classification(name, color)
            Classification._cached_classifications[classification.name] = classification
        return classification

    def __str__(self):
        return f"Classification {self._name} of color {self._color}"
        
    def __repr__(self):
        return f"Classification('{self._name}', {self._color})"
        
    def __eq__(self, other):
        if isinstance(other, Classification):
            return self._name == other._name and self._color == other._color
        return False
    
    def __hash__(self):
        return hash(self._name)



from dataclasses import dataclass


@dataclass(frozen=True)
class ImageChannel:
    """
    Simple data class to store the name and the color of a channel of an image.

    :param name: the name of the channel
    :param color: the RGB color of the channel, with each component between 0 and 1
    """
    name: str
    color: tuple[float, float, float]



from __future__ import annotations
from dataclasses import dataclass


@dataclass(frozen=True)
class PixelLength:
    """
    Simple data class to store pixel size information, along one dimension.

    Can be thought of as the pixel width, pixel height or pixel depth (z-spacing).
    
    :param length: the length of the pixel, by default 1
    :param unit: a text describing the unit of length, by default "pixels"
    """
    length: float = 1.0
    unit: str = 'pixels'

    def is_default(self) -> bool:
        """
        Returns True if this is a default value (length is 1.0 and unit is 'pixels')

        :returns: whether this is a default pixel length
        """
        return self.length == 1.0 and self.unit == 'pixels'

    @staticmethod
    def create_microns(length: float) -> PixelLength:
        """
        Create a PixelLength with a unit of micrometers (µm).

        :param length: the length of the pixel
        :returns: a pixel length of the provided length with the 'micrometer' unit
        """
        return PixelLength(length=length, unit='micrometer')

    @staticmethod
    def create_unknown(length: float) -> PixelLength:
        """
        Create a PixelLength with an unknown unit.

        :param length: the length of the pixel
        :returns: a pixel length of the provided length with no unit
        """
        return PixelLength(length=length, unit=None)


@dataclass(frozen=True)
class PixelCalibration:
    """
    Simple data class for storing pixel calibration information.

    :param length_x: the pixel size along the x-axis
    :param length_y: the pixel size along the y-axis
    :param length_z: the pixel size along the z-axis
    """
    length_x: PixelLength = PixelLength()
    length_y: PixelLength = PixelLength()
    length_z: PixelLength = PixelLength()

    def is_calibrated(self) -> bool:
        """
        Indicate if this PixelCalibration has at least one non-default length.

        :returns: whether this PixelCalibration has at least one non-default length
        """
        for size in [self.length_x, self.length_y, self.length_z]:
            if not size.is_default():
                return True
        return False



import geojson
from typing import Iterable, Union


def geojson_features_from_string(json_string: str, **kwargs) -> Union[Iterable[geojson.Feature], geojson.Feature]:
    """
    Read features from a GeoJSON string.

    If the string encodes a feature collection, the features themselves will be extracted.

    NaNs values are allowed. However, converting a returned feature with a NaN value to a string
    will throw an exception.

    :param json_string: a string representing a GeoJSON
    :param kwargs: additional parameters to pass to the geojson loader
    :returns: a single or a collection of features representing the provided string
    """
    results = _geojson_from_string(json_string, **kwargs)
    if 'features' in results:
        results = results['features']
    return results


def _geojson_from_string(json: str, **kwargs):
    """
    Read a GeoJSON string.
    This is a wrapper around geojson.loads that allows for NaNs by default (and is generally non-strict with numbers).
    """
    # Default parse constant is _enforce_strict_numbers, which fails on NaNs
    if 'parse_constant' in kwargs:
        return geojson.loads(json, **kwargs)
    else:
        return geojson.loads(json, parse_constant=None, **kwargs)



import numpy as np
from typing import Iterable
import shapely
import PIL
from .image_server import ImageServer
from .metadata.image_metadata import ImageMetadata
from .metadata.image_shape import ImageShape
from .metadata.pixel_calibration import PixelCalibration, PixelLength
from .region_2d import Region2D
from ..objects.image_feature import ImageFeature
from ..objects.classification import Classification
from ..objects.draw import draw_geometry


class LabeledImageServer(ImageServer):
    """
    An ImageServer where pixel values are labels corresponding to image features (such as annotations)
    present on an image.

    The returned image will have one timepoint and one z-stack. The size of the remaining dimensions depend
    on the metadata provided when creating the server --- usually, the same as the ImageServer that the labeled image corresponds to.

    The image will only have one resolution level; the downsample for this level may be greater than or less than 1, and consequently region requests and downsamples should be considered relative to the metadata provided at server creation, **not** relative to the downsampled (or upsampled) LabeledImageServer coordinates.
    """

    def __init__(
        self,
        base_image_metadata: ImageMetadata,
        features: Iterable[ImageFeature],
        label_map: dict[Classification, int] = None,
        downsample: float = None,
        multichannel: bool = False,
        resize_method=PIL.Image.Resampling.NEAREST,
        **kwargs,
    ):
        """
        :param base_image_metadata: the metadata of the image containing the image features
        :param features: the image features to draw
        :param label_map: a dictionary mapping a classification to a label. The value of pixels where an image feature with
                          a certain classification is present will be taken from this dictionnary. If not provided, each feature
                          will be assigned a unique integer. All labels must be greater than 0
        :param downsample: the downsample to apply to the image. Can be omitted to use the full resolution image
        :param multichannel: if False, the image returned by this server will have a single channel where pixel values will be unsigned
                             integers representing a label (see the label_map parameter). If True, the number of channels will be
                             equal to the highest label value + 1, and the pixel located at (c, y, x) is a boolean indicating if an annotation
                             with label c is present on the pixel located at (x, y)
        :param resize_method: the resampling method to use when resizing the image for downsampling. Nearest neighbour by default for labeled images.
        :raises ValueError: when a label in label_map is less than or equal to 0
        """
        super().__init__(resize_method=resize_method, **kwargs)

        if label_map is not None and any(label <= 0 for label in label_map.values()):
            raise ValueError(
                "A label in label_map is less than or equal to 0: " + str(label_map)
            )

        self._base_image_metadata = base_image_metadata
        self._downsample = 1 if downsample is None else downsample
        self._multichannel = multichannel
        self._features = [
            f for f in features if label_map is None or f.classification in label_map
        ]
        self._geometries = [
            shapely.affinity.scale(
                shapely.geometry.shape(f.geometry),
                1 / self._downsample,
                1 / self._downsample,
                origin=(0, 0, 0),
            )
            for f in self._features
        ]
        self._tree = shapely.STRtree(self._geometries)

        if label_map is None:
            self._feature_index_to_label = {
                i: i + 1 for i in range(len(self._features))
            }
        else:
            self._feature_index_to_label = {
                i: label_map[self._features[i].classification]
                for i in range(len(self._features))
            }

    def close(self):
        pass

    def _build_metadata(self) -> ImageMetadata:
        return ImageMetadata(
            self._base_image_metadata.path,
            f"{self._base_image_metadata.name} - labels",
            (
                ImageShape(
                    int(self._base_image_metadata.width),
                    int(self._base_image_metadata.height),
                    1,
                    max(self._feature_index_to_label.values(), default=0) + 1
                    if self._multichannel
                    else 1,
                    1,
                ),
            ),
            PixelCalibration(
                PixelLength(
                    self._base_image_metadata.pixel_calibration.length_x.length,
                    self._base_image_metadata.pixel_calibration.length_x.unit,
                ),
                PixelLength(
                    self._base_image_metadata.pixel_calibration.length_y.length,
                    self._base_image_metadata.pixel_calibration.length_y.unit,
                ),
                self._base_image_metadata.pixel_calibration.length_z,
            ),
            False,
            bool if self._multichannel else np.uint32,
            downsamples=[self._downsample],
        )

    def _read_block(self, level: int, region: Region2D) -> np.ndarray:
        if self._multichannel:
            full_image = np.zeros(
                (self.metadata.n_channels, region.height, region.width),
                dtype=self.metadata.dtype,
            )
            feature_indices = self._tree.query(region.geometry)
            labels = set(self._feature_index_to_label.values())

            for label in labels:
                image = PIL.Image.new("1", (region.width, region.height))
                drawing_context = PIL.ImageDraw.Draw(image)

                for i in feature_indices:
                    if label == self._feature_index_to_label[i]:
                        draw_geometry(
                            image.size,
                            drawing_context,
                            shapely.affinity.translate(
                                self._geometries[i], -region.x, -region.y
                            ),
                            1,
                        )
                full_image[label, :, :] = np.asarray(image, dtype=self.metadata.dtype)

            return full_image
        else:
            image = PIL.Image.new("I", (region.width, region.height))
            drawing_context = PIL.ImageDraw.Draw(image)
            for i in self._tree.query(region.geometry):
                draw_geometry(
                    image.size,
                    drawing_context,
                    shapely.affinity.translate(
                        self._geometries[i], -region.x, -region.y
                    ),
                    self._feature_index_to_label[i],
                )
            return np.expand_dims(np.asarray(image, dtype=self.metadata.dtype), axis=0)



import numpy as np
import warnings
import io
import tifffile
from typing import Union
from PIL import Image, ImageCms
from .wrapped_image_server import WrappedImageServer
from .image_server import ImageServer
from .region_2d import Region2D


class IccProfileServer(WrappedImageServer):
    """
    Wrap an ImageServer and apply an ICC Profile to the pixels, if possible.

    If no ICC Profile is provided, an attempt is made to read the profile from the image using PIL.
    This isn't guaranteed to succeed.
    To find out if it was successful, test whether self.icc_transform is not None.

    See http://www.andrewjanowczyk.com/application-of-icc-profiles-to-digital-pathology-images/ 
    for a blog post describing where this may be useful, and providing further code.
    
    Closing this server will close the wrapped server.
    The metadata of this server is equivalent to the metadata of the wrapped server.
    """

    def __init__(
            self,
            base_server: ImageServer,
            icc_profile: Union[bytes, ImageCms.ImageCmsProfile, ImageCms.core.CmsProfile, ImageCms.ImageCmsTransform] = None,
            **kwargs
    ):
        """
        :param base_server: the server to wrap
        :param icc_profile: the ICC profile to apply to the wrapped image server. If omitted, an attempt is made to read the profile from the image.
                            If not successful, a warning will be logged.
        :param resize_method: the resampling method to use when resizing the image for downsampling. Bicubic by default
        """
        super().__init__(base_server, **kwargs)

        if icc_profile is None:
            icc_profile = IccProfileServer._get_icc_bytes_from_path(base_server.metadata.path)

        if icc_profile is None:
            self._icc = None
        elif isinstance(icc_profile, ImageCms.ImageCmsTransform):
            self._icc = icc_profile
        else:
            self._icc = ImageCms.buildTransformFromOpenProfiles(icc_profile, ImageCms.createProfile("sRGB"), "RGB", "RGB")

        if self._icc is None:
            warnings.warn(f'No ICC Profile found for {base_server.metadata.path}. Returning original pixels.')

    @property
    def icc_transform(self) -> ImageCms.ImageCmsTransform:
        """
        Get the transform used to apply the ICC profile.

        If this is None, then the server simply returns the original pixels unchanged.
        """
        return self._icc

    def _read_block(self, level: int, region: Region2D) -> np.ndarray:
        image = self.base_server._read_block(level, region)

        if self._icc:
            return np.transpose(np.array(ImageCms.applyTransform(Image.fromarray(np.transpose(image, axes=[2, 1, 0])), self._icc)), axes=[2, 1, 0])
        else:
            return image
    
    @staticmethod
    def _get_icc_bytes_from_path(path) -> bytes:
        try:
            with tifffile.TiffFile(path) as tif:
                if 34675 in tif.pages[0].tags:
                    return io.BytesIO(tif.pages[0].tags[34675].value)
                else:
                    return None
        except Exception as error:
            warnings.warn(f"Error while retrieving icc profile from {path}: {error}")
            return None



import numpy as np
import dask.array as da
import dask
from dask_image import ndinterp
import warnings
from typing import Union, Iterable
from abc import ABC, abstractmethod
from PIL import Image
from .region_2d import Region2D
from .metadata.image_metadata import ImageMetadata


class ImageServer(ABC):
    """
    An abtract class to read pixels and metadata of an image.

    An image server must be closed (see the close() function) once no longer used.
    """

    def __init__(self, resize_method: Image.Resampling = Image.Resampling.BICUBIC):
        """
        :param resize_method: the resampling method to use when resizing the image for downsampling. Bicubic by default
        """
        super().__init__()
        self._metadata = None
        self._resize_method = resize_method

    @property
    def metadata(self) -> ImageMetadata:
        """
        The image metadata.
        """
        if self._metadata is None:
            self._metadata = self._build_metadata()
        return self._metadata

    def read_region(
        self,
        downsample: float = 1.0,
        region: Union[Region2D, tuple[int, ...]] = None,
        x: int = 0,
        y: int = 0,
        width: int = -1,
        height: int = -1,
        z: int = 0,
        t: int = 0
    ) -> np.ndarray:
        """
        Read pixels from any arbitrary image region, at any resolution determined by the downsample.

        This method can be called in one of two ways: passing a region (as a Region2D object or a tuple of integers),
        or passing x, y, width, height, z and t parameters separately. The latter can be more convenient and readable
        when calling interactively, without the need to create a region object.
        If a region is passed, the other parameters (except for the downsample) are ignored.

        Important: coordinates and width/height values are given in the coordinate space of the full-resolution image,
        and the downsample is applied before reading the region. This means that, except when the downsample is 1.0,
        the width and height of the returned image will usually be different from the width and height passed as parameters.

        :param downsample: the downsample to use
        :param region: a Region2D object or a tuple of integers (x, y, width, height, z, t)
        :param x: the x coordinate of the region to read
        :param y: the y coordinate of the region to read
        :param width: the width of the region to read
        :param height: the height of the region to read
        :param z: the z index of the region to read
        :param t: the t index of the region to read
        :return: a 3-dimensional numpy array containing the requested pixels from the 2D region.
                 The [c, y, x] index of the returned array returns the channel of index c of the
                 pixel located at coordinates [x, y] on the image
        :raises ValueError: when the region to read is not specified
        """
        if region is None:
            region = Region2D(x=x, y=y, width=width, height=height, z=z, t=t)
        elif isinstance(region, tuple):
            # If we have a tuple, use it along with the downsample if available
            region = Region2D(*region)
        if not isinstance(region, Region2D):
            raise ValueError('No valid region provided to read_region method')

        # Fix negative values for width or height
        if region.width < 0 or region.height < 0:
            w = region.width if region.width >= 0 else self.metadata.width - region.x
            h = region.height if region.height >= 0 else self.metadata.height - region.y
            region = Region2D(x=region.x, y=region.y, width=w, height=h, z=region.z, t=region.t)

        level = ImageServer._get_level(self.metadata.downsamples, downsample)
        level_downsample = self.metadata.downsamples[level]
        image = self._read_block(level, region.downsample_region(downsample=level_downsample))
        
        if downsample == level_downsample:
            return image
        else:
            target_size = (round(region.width / downsample), round(region.height / downsample))
            return self._resize(image, target_size, self._resize_method)
        
    def level_to_dask(self, level: int = 0, chunk_width: int = 1024, chunk_height: int = 1024) -> da.Array:
        """
        Return a dask array representing a single resolution of the image.

        Pixels of the returned array can be accessed with the following order:
        (t, c, z, y, x). There may be less dimensions for simple images: for
        example, an image with a single timepoint and a single z-slice will
        return an array of dimensions (c, y, x). However, there will always be
        dimensions x and y, even if they have a size of 1.

        Subclasses of ImageServer may override this function if they can provide
        a faster implementation.

        :param level: the pyramid level (0 is full resolution). Must be less than the number
                      of resolutions of the image
        :param chunk_width: the image will be read chunk by chunk. This parameter specifies the
                            size of the chunks on the x-axis
        :param chunk_height: the size of the chunks on the y-axis
        :returns: a dask array containing all pixels of the provided level
        :raises ValueError: when level is not valid
        """
        if level < 0 or level >= self.metadata.n_resolutions:
            raise ValueError(
                "The provided level ({0}) is outside of the valid range ([0, {1}])".format(level, self.metadata.n_resolutions - 1)
            )

        ts = []
        for t in range(self.metadata.n_timepoints):
            zs = []
            for z in range(self.metadata.n_z_slices):
                xs = []
                for x in range(0, self.metadata.shapes[level].x, chunk_width):
                    ys = []
                    for y in range(0, self.metadata.shapes[level].y, chunk_height):
                        width = min(chunk_width, self.metadata.shapes[level].x - x)
                        height = min(chunk_height, self.metadata.shapes[level].y - y)

                        ys.append(da.from_delayed(
                            dask.delayed(self._read_block)(level, Region2D(x, y, width, height, z, t)),
                            shape=(
                                self.metadata.n_channels,
                                height,
                                width
                            ),
                            dtype=self.metadata.dtype
                        ))
                    xs.append(da.concatenate(ys, axis=1))
                zs.append(da.concatenate(xs, axis=2))
            ts.append(da.stack(zs))
        image = da.stack(ts)

        # Swap channels and z-stacks axis
        image = da.swapaxes(image, 1, 2)        

        # Remove axis of length 1
        axes_to_squeeze = []
        if self.metadata.n_timepoints == 1:
            axes_to_squeeze.append(0)
        if self.metadata.n_channels == 1:
            axes_to_squeeze.append(1)
        if self.metadata.n_z_slices == 1:
            axes_to_squeeze.append(2)
        image = da.squeeze(image, tuple(axes_to_squeeze))
        
        return image
    
    def to_dask(self, downsample: Union[float, Iterable[float]] = None) -> Union[da.Array, tuple[da.Array, ...]]:
        """
        Convert this image to one or more dask arrays, at any arbitary downsample factor.

        It turns out that requesting at an arbitrary downsample level is very slow - currently, all
        pixels are requested upon first compute (even for a small region), and then resized.
        Prefer using ImageServer.level_to_dask() instead.

        :param downsample: the downsample factor to use, or a list of downsample factors to use. If None, all available resolutions will be used
        :return: a dask array or tuple of dask arrays, depending upon whether one or more downsample factors are required
        """

        if downsample is None:
            if self.n_resolutions == 1:
                return self.level_to_dask(level=0)
            else:
                return tuple([self.level_to_dask(level=level) for level in range(self.metadata.n_resolutions)])

        if isinstance(downsample, Iterable):
            return tuple([self.to_dask(downsample=float(d)) for d in downsample])

        level = ImageServer._get_level(self.metadata.downsamples, downsample)
        array = self.level_to_dask(level=level)

        rescale = downsample / self.metadata.downsamples[level]
        input_width = array.shape[-1]
        input_height = array.shape[-2]
        output_width = int(round(input_width / rescale))
        output_height = int(round(input_height / rescale))
        if input_width == output_width and input_height == output_height:
            return array

        # Couldn't find an easy resizing method for dask arrays... so we try this instead
        # TODO: Urgently need something better! Performance is terrible for large images - all pixels requested
        #       upon first compute (even for a small region), and then resized. This is not scalable.
        if array.size > 10000:
            warnings.warn('Warning - calling affine_transform on a large dask array can be *very* slow')

        transform = np.eye(array.ndim)
        transform[array.ndim-1, array.ndim-1] = rescale
        transform[array.ndim-2, array.ndim-2] = rescale
        output_shape = list(array.shape)
        output_shape[-1] = output_width
        output_shape[-2] = output_height

        return ndinterp.affine_transform(array, transform, output_shape=tuple(output_shape))

    @abstractmethod
    def close(self):
        """
        Close this image server.
        
        This should be called whenever this server is not used anymore.
        """
        pass

    @abstractmethod
    def _build_metadata(self) -> ImageMetadata:
        """
        Create metadata for the current image.

        :return: the metadata of the image
        """
        pass

    @abstractmethod
    def _read_block(self, level: int, region: Region2D) -> np.ndarray:
        """
        Read a block of pixels from a specific level.

        Coordinates are provided in the coordinate space of the level, NOT the full-resolution image.
        This means that the returned image should have the width and height specified.
        
        :param level: the pyramidal level to read from
        :param region: the region to read
        :return: a 3-dimensional numpy array containing the requested pixels from the 2D region.
                 The [c, y, x] index of the returned array returns the channel of index c of the
                 pixel located at coordinates [x, y] on the image
        """
        pass
    
    @staticmethod
    def _get_level(all_downsamples: tuple[float], downsample: float, abs_tol=1e-3) -> int:
        """
        Get the level (index) from the image downsamples that is best for fulfilling an image region request.

        This is the index of the entry in self.downsamples that either (almost) matches the requested downsample,
        or relates to the next highest resolution image (so that any required scaling is to reduce resolution).

        :param downsample: the requested downsample value
        :param abs_tol: absolute tolerance when comparing downsample values; this allows for a stored downsample
                        value to be slightly off due to rounding
                        (e.g. requesting 4.0 would match a level 4 +/- abs_tol)
        :return: the level that is best for fulfilling an image region request at the specified downsample
        """
        if len(all_downsamples) == 1 or downsample <= all_downsamples[0]:
            return 0
        elif downsample >= all_downsamples[-1]:
            return len(all_downsamples) - 1
        else:
            # Allow a little bit of a tolerance because downsamples are often calculated
            # by rounding the ratio of image dimensions... and can end up a little bit off
            for level, d in reversed(list(enumerate(all_downsamples))):
                if downsample >= d - abs_tol:
                    return level
            return 0
        
    @staticmethod
    def _resize(image: Union[np.ndarray, Image.Image], target_size: tuple[int, int], resample: int = Image.Resampling.BICUBIC) -> Union[np.ndarray, Image.Image]:
        """
        Resize an image to a target size.

        This uses the implementation from PIL.

        :param image: the image to resize. Either a 3-dimensional numpy array with dimensions (c, y, x)
                      or a PIL image
        :param target_size: target size in (width, height) format
        :param resample: resampling mode to use, by default bicubic
        :return: the resized image, either a 3-dimensional numpy array with dimensions (c, y, x) or a PIL image
        """

        if ImageServer._get_size(image) == target_size:
            return image

        # If we have a PIL image, just resize normally
        if isinstance(image, Image.Image):
            return image.resize(size=target_size, resample=resample)
        # If we have NumPy, do one channel at a time
        else:
            if image.ndim == 2:
                if image.dtype in [np.uint8, np.float32]:
                    pilImage = Image.fromarray(image)
                elif np.issubdtype(image.dtype, np.integer):
                    pilImage = Image.fromarray(image.astype(np.int32), mode='I')
                elif np.issubdtype(image.dtype, np.bool_):
                    pilImage = Image.fromarray(image, "1")
                else:
                    pilImage = Image.fromarray(image.astype(np.float32), mode='F')
                pilImage = ImageServer._resize(pilImage, target_size=target_size, resample=resample)
                return np.asarray(pilImage).astype(image.dtype)
            else:
                return np.stack([
                    ImageServer._resize(image[c, :, :], target_size=target_size, resample=resample) for c in range(image.shape[0])
                ])
        
    @staticmethod
    def _get_size(image: Union[np.ndarray, Image.Image]):
        """
        Get the size of an image as a two-element tuple (width, height).

        :param image: the image whose size should be computed. Either a 3-dimensional numpy array with dimensions (c, y, x)
                      or a PIL image
        """
        return image.size if isinstance(image, Image.Image) else image.shape[1:][::-1]
